#include "util.h"
#include "iptc.h"

#include <stdlib.h>
#include <stdio.h>
#ifndef OS_WIN32
#include <unistd.h>
#endif
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

extern struct prefs_struct prefs;


struct op_label op_labels[]=
{
	{	OP_CHECK,			"check",		"just check embedded data structures"	},
	{	OP_DUMP,			"dump",			"show embedded data (headers only)"	},
	{	OP_DUMP_FULL,		"dump-full",	"show all embedded data (including values)"	},
	{	OP_FILTER,			"filter",		"filter embedded data (use -i OR -e)"	},
	/* TODO:
	   - purge (iptc, exif, .., *),
	   - add (IPTC.0xnnnn.value, Exif..),
	   - remove? -> filter,
	   - merge (picture file+iptc/exif/photoshop file)
	   - extract (exif)
	*/
	/* reserved */
	{	OP_EOT, NULL, NULL	}
};

struct log_level_descr log_levels[]=
{
	{	LOG_LEVEL_QUIET,	"quiet, no output at all"	},
	{	LOG_LEVEL_ERRORS,	"error messages only (default)"	},
	{	LOG_LEVEL_WARNINGS,	"warning and error messages"	},
	{	LOG_LEVEL_INFO,		"informative, warning and error messages"	},
	{	LOG_LEVEL_DEBUG,	"all messages including debug ones"	},
	/* reserved */
	{	LOG_LEVEL_EOT,		NULL	}
};

struct exit_code_label exit_code_labels[]=
{
	{	EXIT_CODE_NORMAL,				"normal exit"	},
	{	EXIT_CODE_USAGE_ERROR,			"usage error"	},
	{	EXIT_CODE_ASYNCHRONOUS_SIGNAL,	"asynchronous signal termination"	},
	{	EXIT_CODE_NORMAL_WITH_WARNINGS,	"normal exit, with warning(s)"	},
	{	EXIT_CODE_NORMAL_WITH_ERRORS,	"normal exit, with error(s)"	},
	{	EXIT_CODE_FATAL_ERROR,			"fatal error encountered"	},
	/* reserved */
	{	EXIT_EOT, NULL	}
};

unsigned char* rewrite_cache;
unsigned long int rewrite_cache_offset;
unsigned long int rewrite_cache_size;
FILE* output_file;
char* output_filename;
size_t input_length;


size_t fsize(FILE* file)
{
	size_t tmp;
	fseek(file, 0, SEEK_END);
	tmp=(size_t)ftell(file);
	fseek(file, 0, SEEK_SET);
	return(tmp);
}

bool fexist(const char* filename)
{
	struct stat st;

	if (stat(filename, &st)==0)
		return(true);
	return(false);
}

int frename(const char* source_filename, const char* target_filename)
{
	int err;

	if (fexist(target_filename))
	{
#ifdef OS_WIN32
		err=remove(target_filename);
		if (err!=0)
			error("couldn't remove '%s' (error #%d)",
					target_filename, err);
#else
		err=unlink(target_filename);
		if (err!=0)
			error("couldn't remove '%s' (error #%d)",
					target_filename, err);
#endif
	}

	err=rename(source_filename, target_filename);
	return(err);
}

bool fisfile(const char* filename)
{
	struct stat st;

	if (stat(filename, &st)<0)
		return(false);
	if (S_ISREG(st.st_mode))
		return(true);
	return(false);
}

bool fislink(const char* filename)
{
	struct stat st;

	if (stat(filename, &st)<0)
		return(false);
	if (S_ISLNK(st.st_mode))
		return(true);
	return(false);
}

bool fisdir(const char* filename)
{
	struct stat st;

	if (stat(filename, &st)<0)
		return(false);
	if (S_ISDIR(st.st_mode))
		return(true);
	return(false);
}

char* fextension(const char* path)
/* return a newly allocated string with the file extension if any
  (or "" if none). it's up to the caller to free the returned pointer */
{
	char* extension;
	const char* filename;
	int dot;
	int sep;

	/* keep the filename only */
	sep=strrpos(path, '/');
	if (sep==0)
		filename=path;
	else
		filename=path+sep;

	dot=strrpos(filename, '.');
	if (dot==0)
		extension=strdup("");
	else
		extension=strdup(filename+dot);
	return(extension);
}

char* fbasename(const char* path)
/* return a newly allocated string with the file extension if any
  (or "" if none). it's up to the caller to free the returned pointer */
{
	char* basename;
	const char* filename;
	int dot;
	int sep;

	/* keep the filename only */
	sep=strrpos(path, '/');
	if (sep==0)
		filename=path;
	else
		filename=path+sep;

	dot=strrpos(filename, '.');
	if (dot==0)
		basename=strdup(filename);
	else
		basename=strndup(filename, dot-1);
	return(basename);
}

char* ffilename(const char* path)
/* return a newly allocated string with the file name if any
  (or "" if none). it's up to the caller to free the returned pointer */
{
	char* filename;
	int sep;

	sep=strrpos(path, '/');
	if (sep==0)
		filename=strdup(path);
	else
		filename=strdup(path+sep);
	return(filename);
}

char* fdirname(const char* path)
/* return a newly allocated string with the dirname (everything but the filename) if any
  (or "" if none). it's up to the caller to free the returned pointer */
{
	char* dirname;
	int sep;

	sep=strrpos(path, '/');
	if (sep==0)
		dirname=strdup("");
	else
		dirname=strndup(path, sep-1);
	return(dirname);
}

char* fcleanpath(const char* path)
/* return a newly allocated string with the clean path (no duplicate or trailing /).
   it's up to the caller to free the returned pointer */
{
	const char* ptr;
	size_t len;

	/* skip trailing slashes */
	len=strlen(path);
	ptr=path+len-1;
	while ((ptr>=path) && (*ptr=='/'))
	{
		ptr--;
		len--;
	}
	return(strndup(path, len));
}

unsigned char getbyte(unsigned char* ptr, unsigned long int offset)
{
	return(ptr[offset]);
}

unsigned short int getword(unsigned char* ptr, unsigned long int offset)
{
	return((ptr[offset]<<8)|ptr[offset+1]);
}

unsigned long int getlong(unsigned char* ptr, unsigned long int offset)
{
	return((((ptr[offset]<<8)|ptr[offset+1])<<16)|((ptr[offset+2]<<8)|ptr[offset+3]));
}


unsigned char* getstring(unsigned char* ptr, unsigned long int offset, unsigned char* buf,
		const unsigned long int len)
{
	register unsigned int i;

	for (i=0; i<len; i++)
		buf[i]=ptr[offset+i];
	buf[i]='\0';
	return(buf);
}

unsigned char* getraw(unsigned char* ptr, unsigned long int offset, unsigned char* buf,
		const unsigned long int len)
{
	memcpy(buf, ptr+offset, len*sizeof(unsigned char));
	return(buf);
}

void setbyte(unsigned char* ptr, unsigned long int offset, unsigned char byte)
{
	ptr[offset]=byte;
}

void setword(unsigned char* ptr, unsigned long int offset, unsigned short int word)
{
	ptr[offset]=(word&0xff00)>>8;
	ptr[offset+1]=word&0x00ff;
}

void setlong(unsigned char* ptr, unsigned long int offset, unsigned long int dword)
{
	ptr[offset]=(dword&0xff000000UL)>>24;
	ptr[offset+1]=(dword&0x00ff0000UL)>>16;
	ptr[offset+2]=(dword&0x0000ff00UL)>>8;
	ptr[offset+3]=dword&0x000000ffUL;
}

void message_fatal_error(char* format, va_list args)
{
	if (prefs.log_level>=LOG_LEVEL_ERRORS)
	{
		context_print_text();
		prefs.context_errors++;
		prefs.errors++;
		fprintf(stderr, "error: ");
		vfprintf(stderr, format, args);
		fprintf(stderr, "\nfatal error encountered, bailing out\n");
		context_print_info();
	}
	_exit(EXIT_CODE_FATAL_ERROR);
}

void message_error(char* format, va_list args)
{
	if (prefs.log_level>=LOG_LEVEL_ERRORS)
	{
		context_print_text();
		prefs.context_errors++;
		prefs.errors++;
		fprintf(stderr, "error: ");
		vfprintf(stderr, format, args);
		fprintf(stderr, "\n");
	}
}

void message_warning(char* format, va_list args)
{
	if (prefs.log_level>=LOG_LEVEL_WARNINGS)
	{
		context_print_text();
		prefs.context_warnings++;
		prefs.warnings++;
		fprintf(stderr, "warning: ");
		vfprintf(stderr, format, args);
		fprintf(stderr, "\n");
	}
}

void message_info(char* format, va_list args)
{
	if (prefs.log_level>=LOG_LEVEL_INFO)
	{
		fprintf(stdout, "info: ");
		vfprintf(stdout, format, args);
		fprintf(stdout, "\n");
		fflush(stdout);
	}
}

void message_debug(char* format, va_list args)
{
	if (prefs.log_level>=LOG_LEVEL_DEBUG)
	{
		fprintf(stdout, "debug: ");
		vfprintf(stdout, format, args);
		fprintf(stdout, "\n");
		fflush(stdout);
	}
}

void fatal_error(char* format, ...)
{
	va_list args;

	va_start(args, format);
	message_fatal_error(format, args);
	va_end(args);
}

void error(char* format, ...)
{
	va_list args;

	va_start(args, format);
	message_error(format, args);
	va_end(args);
}

void warning(char* format, ...)
{
	va_list args;

	va_start(args, format);
	message_warning(format, args);
	va_end(args);
}

void info(char* format, ...)
{
	va_list args;

	va_start(args, format);
	message_info(format, args);
	va_end(args);
}

void debug(char* format, ...)
{
	va_list args;

	va_start(args, format);
	message_debug(format, args);
	va_end(args);
}

void message(enum SEVERITY severity, char* format, ...)
{
	va_list args;

	va_start(args, format);
	switch (severity)
	{
		case SEV_FATAL_ERROR:
			message_fatal_error(format, args);
			break;

		case SEV_ERROR:
			message_error(format, args);
			break;

		case SEV_WARNING:
			message_warning(format, args);
			break;

		case SEV_INFO:
			message_info(format, args);
			break;

		case SEV_DEBUG:
			message_debug(format, args);
			break;
	}
	va_end(args);
}

void context_set_text(char* text)
{
	prefs.context_text=text;
}

void context_print_text(void)
{
	/* show context information when first warning/error is encounter */
	if (prefs.context_text)
		if ((prefs.context_warnings+prefs.context_errors)==0)
			fprintf(stderr, "in file '%s':\n", prefs.context_text);
}

void context_print_info(void)
{
	if ((prefs.context_warnings+prefs.context_errors)>0)
		fprintf(stderr, "%lu warning(s), %lu error(s)\n",
				prefs.context_warnings, prefs.context_errors);
	if (prefs.fix && (prefs.context_fixes>0))
		fprintf(stderr, "%lu fix(es)\n",
				prefs.fixes);
}

void context_reset(void)
{
	prefs.context_text=NULL;
	prefs.context_warnings=0;
	prefs.context_errors=0;
	prefs.context_fixes=0;
}

struct op_label* op_match_label(char* label)
{
	register int i;

	i=0;
	while (op_labels[i].op!=OP_EOT)
	{
		if (strcmp(op_labels[i].label, label)==0)
			return(&op_labels[i]);
		i++;
	}
	return(NULL);
}

void init_rewrite(void)
{
	char* filename;
	size_t len;
	int fd;
	FILE* file;

	if (!prefs.rewrite)
		fatal_error("init_rewrite: rewrite flag not set");
	if (output_file)
		fatal_error("init_rewrite: bad output file handler");
	if (!output_filename)
		fatal_error("init_rewrite: bad output filename");

	/* get a temporary filename */
	len=strlen(output_filename);
	filename=malloc((len+1+6+1)*sizeof(char));
	if (!filename)
		fatal_error("couldn't allocate space for init_rewrite");
	strcpy(filename, output_filename);
	strcat(filename, ".XXXXXX");
	fd=mkstemp(filename);
	if (fd<0)
		fatal_error("couldn't create a temporary file from '%s'", output_filename);

	/* open file for writing */
	file=fdopen(fd, "w+b");
	if (!file)
	{
		close(fd);
		fatal_error("coulnd't open temporary file '%s' for writing", output_filename);
	}

	/* write back values to caller */
	output_file=file;
	free(output_filename); /* free previous allocation (from caller) */
	output_filename=strdup(filename);
	free(filename);

	/* cached rewriting */
	if (prefs.rewrite_cached)
	{
		rewrite_cache=malloc(input_length*sizeof(char));
		if (!rewrite_cache)
			fatal_error("couldn't allocate space for rewrite cache (%lu byte(s))",
					rewrite_cache_size);
		rewrite_cache_offset=0;
		rewrite_cache_size=input_length;
	}
}

size_t dump_rewrite(unsigned char* buffer, unsigned long int length)
/* returns the number of bytes written */
{
	size_t len;

	if (!prefs.rewrite)
		fatal_error("init_rewrite: rewrite flag not set");

	debug("dump_rewrite: writing %lu byte(s)", length);
	len=0;
	if (length>0)
	{
		if (prefs.rewrite_cached)
		{
			if (rewrite_cache_offset+length>rewrite_cache_size)
			{
				/* reallocate more space for cached writing */
				rewrite_cache_size+=length+1024;
				rewrite_cache=realloc(rewrite_cache, rewrite_cache_size);
				if (!rewrite_cache)
					fatal_error("couldn't rellocate more space for rewrite_cache (%lu byte(s))",
							rewrite_cache_size);
			}
			/* write bytes to cache */
			memcpy(rewrite_cache+rewrite_cache_offset, buffer, length);
			rewrite_cache_offset+=length;
			len=length;
		}
		else
		{
			/* dump bytes to file */
			len=fwrite(buffer, 1, length, output_file);
			if (len!=length)
				error("dump_rewrite: error %d while attempting to write %lu byte(s) (%lu written)",
						errno, length, len);
		}
	}
	return(len);
}

void post_rewrite(unsigned char* buffer, struct parser_result* result)
{
	unsigned long int len;

	if (!prefs.rewrite)
		fatal_error("init_rewrite: rewrite flag not set");
	if (!result)
		fatal_error("init_rewrite: bad pointer to struct result");

	/* write remaining bytes */
	if (input_length>result->parsed_bytes)
	{
		len=input_length-result->parsed_bytes;
		/*
		if (result->parsed_bytes!=result->written_bytes)
			len-=(result->parsed_bytes-result->written_bytes);
		*/
	}
	else
		len=0;
	if (prefs.rewrite_cached)
	{
		if (len>0)
		{
			debug("post_rewrite: writing remaining %lu byte(s) to cache", len);
			memcpy(rewrite_cache+rewrite_cache_offset, buffer+result->parsed_bytes, len);
			rewrite_cache_offset+=len;
		}
		/* flush cache to disk */
		len=fwrite(rewrite_cache, 1, rewrite_cache_offset, output_file);
		if (len!=rewrite_cache_offset)
			error("post_rewrite: error %d while attempting to flush %lu cached bytes (%lu written)",
					errno, rewrite_cache_offset, len);
		if (rewrite_cache_offset>input_length)
			info("rewritten file is %lu byte(s) bigger than the original",
					rewrite_cache_offset-input_length);
		else
			if (rewrite_cache_offset<input_length)
				info("rewritten file is %lu byte(s) smaller than the original",
						input_length-rewrite_cache_offset);
			else
				info("rewritten file and original file have the same size");
	}
	else
		if (len>0)
		{
			debug("post_rewrite: writing remaining %lu byte(s)", len);
			dump_rewrite(buffer+result->parsed_bytes, len);
		}
}

bool get_unsigned_value(const char* param, unsigned long int* var)
{
	unsigned int value;
	char* chk;

	value=(unsigned int) strtoul(param, &chk, 10);
	if (!*chk)
	{
		*var=value;
		return(true);
	}
	return(false);
}

bool get_signed_value(const char* param, long int* var)
{
	int value;
	char* chk;

	value=(int)strtol(param, &chk, 10);
	if (!*chk)
	{
		*var=value;
		return(true);
	}
	return(false);
}

bool get_hexa_value(const char* param, unsigned long int* var)
{
	unsigned int value;
	char* chk;

	value=strtoul(param, &chk, 16);
	if (!*chk)
	{
		*var=value;
		return(true);
	}
	return(false);
}

size_t strpos(const char* string, const char c)
{
	char* ptr;

	ptr=(char*)string;
	while (*ptr)
		if (*ptr==c)
			return((size_t)(ptr-string+1));
		else
			ptr++;
	return(0);
}

size_t strrpos(const char* string, const char c)
{
	char* ptr;

	ptr=(char*)string+strlen(string)-1;
	while (ptr>=string)
		if (*ptr==c)
			return((size_t)(ptr-string+1));
		else
			ptr--;
	return(0);
}

void filter_load(FILE* list_file)
{
	char* ptr;
	char* p;
	ssize_t len;
	unsigned long int line;
	unsigned long int errors;
	char* buffer;
	size_t size;

	if (!list_file)
		error("filter_load: bad input file handler");

	iptc_filter_table_init();
	line=0;
	errors=0;
	buffer=NULL;
	size=100;
	while ((len=getline(&buffer, &size, list_file))!=-1)
	{
		buffer[len]='\0';
		debug("filter_load: line %lu, %lu byte(s)", line, len);

		ptr=buffer;
		/* skip leading whitespaces if any */
		while ((*ptr==' ') || (*ptr=='\t'))
			ptr++;

		/* skip trailing whitespaces if any */
		p=buffer+len-1;
		while ((p>=ptr) && ((*p==' ') || (*p=='\t') || (*p=='\n')))
			*p--='\0';

		/* ignore empty lines as well as comments */
		if (!*ptr || (*ptr=='#'))
			debug("filter_load: ignore blank line or comment at line %lu", line);
		else
		/* IPTC filter masks decoding */
		if (strncmp(ptr, "IPTC.", 5)==0)
		{
			/* possible IPTC filter masks:
				IPTC.*					all
				IPTC.0xnnnn				record+dataset (4-digit hexadecimal)
				IPTC.0xnnnn-0xnnnn		record+dataset range (4-digit hexadecimals)
				IPTC.nnn:nnn			record+dataset (1-3 digits per decimal value, each value [0-255])
				IPTC.nnn:*				all datasets from record nnn (1-3 digits per decimal value, each value [0-255])
				IPTC.nnn:nnn-nnn		record and dataset range (1-3 digits per decimal value, each value [0-255])
			   for instance:
				IPTC.*
				IPTC.0x0219
				IPTC.0x0200-0x07ff
				IPTC.2:*
				IPTC.2:25
				IPTC.2:0-255
			*/
			int pos;

			debug("filter_load: found IPTC filter item");
			ptr+=5;
			if (strcmp(ptr, "*")==0)
				iptc_filter_add_range(0x0000, 0xffff);
			else
			if (strncmp(ptr, "0x", 2)==0)
			{
				/* hexadecimal */
				unsigned long int code;
				unsigned long int code2;

				p=strstr(ptr, "-0x");
				if (p)
				{
					/* range value */
					debug("filter_load: hexadecimal range at line %lu", line);
					*p='\0';
					/* skip the range separator */
					p++;
					if ((strlen(ptr)==6) && get_hexa_value(ptr, &code)
						&& (strlen(p)==6) && get_hexa_value(p, &code2))
						iptc_filter_add_range(code, code2);
					else
					{
						errors++;
						error("filter_load: bad hexadecimal range (0xhhhh-0xhhhh) at line %lu",
								line);
					}
				}
				else
				{
					/* unary value */
					debug("filter_load: hexadecimal value at line %lu", line);
					if ((strlen(ptr)==6) && get_hexa_value(ptr, &code))
						iptc_filter_add_value(code);
					else
					{
						errors++;
						error("filter_load: bad hexadecimal value (0xhhhh) at line %lu",
								line);
					}
				}
			}
			else
			{
				/* decimal duplet */
				unsigned long int record;
				unsigned long int dataset;
				unsigned long int dataset2;

				pos=strpos(ptr, ':');
				if (pos)
				{
					debug("filter_load: decimal value at line %lu", line);
					ptr[pos-1]='\0';
					p=ptr+pos;

					/* record value */
					if (get_unsigned_value(ptr, &record) && (record<=0xff))
					{
						if (strcmp(p, "*")==0)
							iptc_filter_add_range((unsigned short int)(record<<8),
									(unsigned short int)(record<<8)|0xff);
						else
						{
							/* range value */
							ptr=p;
							pos=strpos(ptr, '-');
							if (pos)
							{
								ptr[pos-1]='\0';
								p=ptr+pos;

								if (get_unsigned_value(ptr, &dataset) && (dataset<=0xff)
									&& get_unsigned_value(p, &dataset2) && (dataset2<=0xff))
									iptc_filter_add_range((unsigned short int)(record<<8)|dataset,
											(unsigned short int)(record<<8)|dataset2);
								else
								{
									errors++;
									error("filter_load: bad range value in decimal duplet (nnn-nnn) at line %lu",
											line);
								}
							}
							else
							{
								/* unary value */
								if (get_unsigned_value(p, &dataset) && (dataset<0xff))
									iptc_filter_add_value((unsigned short int)((record<<8)|dataset));
								else
								{
									errors++;
									error("filter_load: bad dataset value in decimal duplet (nnn) at line %lu",
											line);
								}
							}
						}
					}
					else
					{
						errors++;
						error("filter_load: bad record value in decimal duplet (nnn) at line %lu",
								line);
					}
				}
				else
				{
					errors++;
					error("filter_load: bad decimal duplet value (nnn:nnn) at line %lu: %s",
							line, ptr);
				}
			}
		}
/*		else
		if (strncmp(buffer, "Exif.", 5))
		{
		}
		else
		if (strncmp(buffer, "Adobe.", 6))
		{
		}
*/		else
		{
			errors++;
			error("filter_load: illegal filter file contents at line %lu (%lu byte(s))",
					line, len);
			debug("filter_load: line is: %s", buffer);
		}
		line++;
	}
	free(buffer);
	debug("filter_load: %lu line(s), %lu error(s)", line, errors);
	if (errors)
		fatal_error("filter_load: bad filter file");
}

void backup_file(const char* filename)
{
	char* backup_filename;
	int err;

	/* build the backup filename */
	backup_filename=malloc((strlen(filename)+1+4)*sizeof(char));
	if (!backup_filename)
		fatal_error("couldn't allocate space for backup filename");
	strcpy(backup_filename, filename);
	strcat(backup_filename, ".bak");
	debug("backing up '%s' to '%s'", filename, backup_filename);

	/* remove any already existing backup file */
	if (fisfile(backup_filename) || fislink(backup_filename))
	{
		err=remove(backup_filename);
		if (err!=0)
			error("couldn't remove previous backup file '%s' (error #%d)",
					backup_filename, err);
	}

	/* rename the original file to back it up */
	if (fisfile(filename) || fislink(backup_filename))
	{
		err=frename(filename, backup_filename);
		if (err!=0)
			error("couldn't create backup '%s' from '%s' (error #%d)",
				backup_filename, filename, err);
	}

	/* free resources */
	free(backup_filename);
	backup_filename=NULL;
}

void dump_hexa(unsigned char* value, unsigned long int length)
{	
	unsigned long int i;

	printf("[%lu byte(s)]:", length);
	if (prefs.wrap_dump>0)
		printf("\n");
	else
		printf(" ");
	for (i=1; i<=length; i++)
	{
		printf("%02x ", value[i-1]);
		if (prefs.wrap_dump>0)
			/* append a newline char each 32 hexabyte printed */
			if ((i%prefs.wrap_dump)==0)
				printf("\n");
	}
	if (prefs.wrap_dump>0)
	{
		if (length!=prefs.wrap_dump) /* newline if not just appended */
			printf("\n");
	}
	else
		printf("\n");
}
